' DES Encryption Implementation by: Chris Brown(2007) aka Zamaster
' Compiles latest CVS in .17 with modification by cha0s
'
' Modified by: Nicolas Mougin aka mougino in 2016 to support CBC
' (Cypher Block Chaining) with an Initialization Vector IV, making
' each cypher block dependent on all the previous Message blocks
' through an initial XOR operation (stronger security).
'
' -Takes plain text and DES encrypts it
' -Keys must be 64 bits in size, or 16 hex places/8 char places
' -IVs must be 64 bits in size, or 16 hex places/8 char places
' -Set ed in DES_Encrypt to 2 for decryption, 1 for encryption
'

DECLARE FUNCTION DES_Encrypt  (BYREF mtext AS STRING, BYVAL mkey AS ULONGINT, BYVAL IV AS ULONGINT, BYVAL ed AS UINTEGER) AS STRING
DECLARE FUNCTION DES_BitCopy  (BYVAL tonum AS UINTEGER, BYVAL toval AS ULONGINT, BYVAL tobits AS UINTEGER, BYVAL fromnum AS UINTEGER, BYVAL fromval AS ULONGINT, BYVAL frombits AS UINTEGER) AS ULONGINT   
DECLARE SUB      DES_Split56  (BYVAL mk AS ULONGINT, BYREF lh AS ULONGINT, BYREF rh AS ULONGINT)
DECLARE SUB      DES_Split64  (BYVAL mk AS ULONGINT, BYREF lh AS ULONGINT, BYREF rh AS ULONGINT)
DECLARE FUNCTION DES_RotateL56(BYVAL wkey AS ULONGINT, BYVAL plcs AS UINTEGER) AS ULONGINT
DECLARE FUNCTION DES_StrToHex (BYREF convstr AS STRING) AS STRING
DECLARE FUNCTION DES_HexToStr (BYREF convstr AS STRING) AS STRING

FUNCTION DES_Encrypt(BYREF mtext AS STRING, BYVAL mkey AS ULONGINT, BYVAL IV AS ULONGINT, BYVAL ed AS UINTEGER) AS STRING
    DIM AS UINTEGER i, a, b, t, z, div8
    DIM AS ULONGINT Kpi, mpc, mpi

    ' Concatenate the Message and a Padding String to have a length multiple of 8 octets
    DIM AS LONG PS_val = 8 - (LEN(mtext) MOD 8)
    DIM AS STRING PS
    FOR i = 1 TO PS_val
        PS += CHR(PS_val)
    NEXT
    DIM AS STRING pt = DES_StrToHex(mtext + PS) ' pt = hexadecimal-encoded Message
    div8 = LEN(pt) \ 16 ' number of 64-bit blocks (8-octet blocks) in the Message
    
    ' Step 1: Create 16 subkeys, PC1 to PC16, each of which is 48-bits long
    '--------------------------------------------------------------------------------------------------

    STATIC AS UINTEGER PC1(1 TO 56) = _
    { 57, 49, 41, 33, 25, 17,  9, _
       1, 58, 50, 42, 34, 26, 18, _
      10,  2, 59, 51, 43, 35, 27, _
      19, 11,  3, 60, 52, 44, 36, _
      63, 55, 47, 39, 31, 23, 15, _
       7, 62, 54, 46, 38, 30, 22, _
      14,  6, 61, 53, 45, 37, 29, _
      21, 13,  5, 28, 20, 12,  4 }
    FOR i = 1 TO 56
        Kpi = DES_BitCopy(i,Kpi,56,PC1(i),mkey,64)
    NEXT i
   
    DIM AS ULONGINT LKpi(0 TO 16), RKpi(0 TO 16), FKpi(1 TO 16, 1 TO 2)
    DES_Split56 Kpi, LKpi(0), RKpi(0)

    STATIC AS INTEGER ShiftSchedule(1 TO 16) = { 1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1 }
    FOR i = 1 TO 16
        READ b
        LKpi(i) = DES_RotateL56(LKpi(i-1),ShiftSchedule(i))
        RKpi(i) = DES_RotateL56(RKpi(i-1),ShiftSchedule(i))
    NEXT i
    FOR i = 1 TO 16
        FKpi(i,1) = (LKpi(i) SHL 28) OR RKpi(i)
    NEXT i

    STATIC AS INTEGER PC2(1 TO 48) = _
    { 14, 17, 11, 24,  1,  5, _
       3, 28, 15,  6, 21, 10, _
      23, 19, 12,  4, 26,  8, _
      16,  7, 27, 20, 13,  2, _
      41, 52, 31, 37, 47, 55, _
      30, 40, 51, 45, 33, 48, _
      44, 49, 39, 56, 34, 53, _
      46, 42, 50, 36, 29, 32 }
    FOR a = 1 TO 16
        FOR i = 1 TO 48
            FKpi(a,2) = DES_BitCopy(i,FKpi(a,2),48,PC2(i),FKpi(a,1),56)
        NEXT i
    NEXT a
    DIM AS STRING   CypherText
    DIM AS ULONGINT LMpi(1 TO 2), RMpi(1 TO 2), expr(1 TO 2), bitgrab, xbit, ybit, nbit, fbit
    DIM AS UINTEGER shiftbits
   
    ' Step 2: Encode each 64-bit block of data with initial permutation IP
    '--------------------------------------------------------------------------------------------------

    STATIC AS UINTEGER SBox(1 TO 8, 0 TO 15, 0 TO 3) = _
    { { { 14,  0,  4, 15 }, {  4, 15,  1, 12 }, { 13,  7, 14,  8 }, {  1,  4,  8,  2 }, {  2, 14, 13,  4 }, { 15,  2,  6,  9 }, { 11, 13,  2,  1 }, {  8,  1, 11,  7 }, {  3, 10, 15,  5 }, { 10,  6, 12, 11 }, {  6, 12,  9,  3 }, { 12, 11,  7, 14 }, {  5,  9,  3, 10 }, {  9,  5, 10,  0 }, {  0,  3,  5,  6 }, {  7,  8,  0, 13 } }, _
      { { 15,  3,  0, 13 }, {  1, 13, 14,  8 }, {  8,  4,  7, 10 }, { 14,  7, 11,  1 }, {  6, 15, 10,  3 }, { 11,  2,  4, 15 }, {  3,  8, 13,  4 }, {  4, 14,  1,  2 }, {  9, 12,  5, 11 }, {  7,  0,  8,  6 }, {  2,  1, 12,  7 }, { 13, 10,  6, 12 }, { 12,  6,  9,  0 }, {  0,  9,  3,  5 }, {  5, 11,  2, 14 }, { 10,  5, 15,  9 } }, _
      { { 10, 13, 13,  1 }, {  0,  7,  6, 10 }, {  9,  0,  4, 13 }, { 14,  9,  9,  0 }, {  6,  3,  8,  6 }, {  3,  4, 15,  9 }, { 15,  6,  3,  8 }, {  5, 10,  0,  7 }, {  1,  2, 11,  4 }, { 13,  8,  1, 15 }, { 12,  5,  2, 14 }, {  7, 14, 12,  3 }, { 11, 12,  5, 11 }, {  4, 11, 10,  5 }, {  2, 15, 14,  2 }, {  8,  1,  7, 12 } }, _
      { {  7, 13, 10,  3 }, { 13,  8,  6, 15 }, { 14, 11,  9,  0 }, {  3,  5,  0,  6 }, {  0,  6, 12, 10 }, {  6, 15, 11,  1 }, {  9,  0,  7, 13 }, { 10,  3, 13,  8 }, {  1,  4, 15,  9 }, {  2,  7,  1,  4 }, {  8,  2,  3,  5 }, {  5, 12, 14, 11 }, { 11,  1,  5, 12 }, { 12, 10,  2,  7 }, {  4, 14,  8,  2 }, { 15,  9,  4, 14 } }, _
      { {  2, 14,  4, 11 }, { 12, 11,  2,  8 }, {  4,  2,  1, 12 }, {  1, 12, 11,  7 }, {  7,  4, 10,  1 }, { 10,  7, 13, 14 }, { 11, 13,  7,  2 }, {  6,  1,  8, 13 }, {  8,  5, 15,  6 }, {  5,  0,  9, 15 }, {  3, 15, 12,  0 }, { 15, 10,  5,  9 }, { 13,  3,  6, 10 }, {  0,  9,  3,  4 }, { 14,  8,  0,  5 }, {  9,  6, 14,  3 } }, _
      { { 12, 10,  9,  4 }, {  1, 15, 14,  3 }, { 10,  4, 15,  2 }, { 15,  2,  5, 12 }, {  9,  7,  2,  9 }, {  2, 12,  8,  5 }, {  6,  9, 12, 15 }, {  8,  5,  3, 10 }, {  0,  6,  7, 11 }, { 13,  1,  0, 14 }, {  3, 13,  4,  1 }, {  4, 14, 10,  7 }, { 14,  0,  1,  6 }, {  7, 11, 13,  0 }, {  5,  3, 11,  8 }, { 11,  8,  6, 13 } }, _
      { {  4, 13,  1,  6 }, { 11,  0,  4, 11 }, {  2, 11, 11, 13 }, { 14,  7, 13,  8 }, { 15,  4, 12,  1 }, {  0,  9,  3,  4 }, {  8,  1,  7, 10 }, { 13, 10, 14,  7 }, {  3, 14, 10,  9 }, { 12,  3, 15,  5 }, {  9,  5,  6,  0 }, {  7, 12,  8, 15 }, {  5,  2,  0, 14 }, { 10, 15,  5,  2 }, {  6,  8,  9,  3 }, {  1,  6,  2, 12 } }, _
      { { 13,  1,  7,  2 }, {  2, 15, 11,  1 }, {  8, 13,  4, 14 }, {  4,  8,  1,  7 }, {  6, 10,  9,  4 }, { 15,  3, 12, 10 }, { 11,  7, 14,  8 }, {  1,  4,  2, 13 }, { 10, 12,  0, 15 }, {  9,  5,  6, 12 }, {  3,  6, 10,  9 }, { 14, 11, 13,  0 }, {  5,  0, 15,  3 }, {  0, 14,  3,  5 }, { 12,  9,  5,  6 }, {  7,  2,  8, 11 } } }
    CypherText = ""

    FOR z = 0 TO div8-ed ' for each 64-bit block (8-octet block) of the Message
       
        ' CBC (Cypher Block Chaining) mode: when DEcrypting, starting at the second
        ' block, take the value of last cypher block as next Initialization Vector
        IF ed = 2 AND z > 0 THEN IV = mpc

        ' Get LONG value of the z'th 64-bit block of the Message
        mpc = VALLNG("&H" + MID$(pt, z*16+1, 16))

        ' CBC (Cypher Block Chaining) mode: when ENcrypting,
        ' XOR the clear block with the Initialization Vector
        IF ed = 1 THEN mpc = mpc XOR IV

        mpi = 0
        STATIC AS UINTEGER IP(1 TO 64) = { 58, 50, 42, 34, 26, 18, 10,  2, _
                                           60, 52, 44, 36, 28, 20, 12,  4, _
                                           62, 54, 46, 38, 30, 22, 14,  6, _
                                           64, 56, 48, 40, 32, 24, 16,  8, _
                                           57, 49, 41, 33, 25, 17,  9,  1, _
                                           59, 51, 43, 35, 27, 19, 11,  3, _
                                           61, 53, 45, 37, 29, 21, 13,  5, _
                                           63, 55, 47, 39, 31, 23, 15,  7 }

        FOR i = 1 TO 64
            mpi = DES_BitCopy(i, mpi, 64, IP(i), mpc, 64)
        NEXT i
        DES_Split64 mpi, LMpi(1), RMpi(1)
       
        FOR i = 1 TO 16
            LMpi(2) = RMpi(1)   
   
            expr(1) = RMpi(1)

            STATIC AS UINTEGER ExpansionSet(1 TO 48) = { 32,  1,  2,  3,  4,  5, _
                                                          4,  5,  6,  7,  8,  9, _
                                                          8,  9, 10, 11, 12, 13, _
                                                         12, 13, 14, 15, 16, 17, _
                                                         16, 17, 18, 19, 20, 21, _
                                                         20, 21, 22, 23, 24, 25, _
                                                         24, 25, 26, 27, 28, 29, _
                                                         28, 29, 30, 31, 32,  1 }


            expr(2) = 0
            FOR t = 1 TO 48
                expr(2) = DES_BitCopy(t,expr(2),48,ExpansionSet(t),expr(1),32)
            NEXT t
            SELECT CASE ed
            CASE 1
                expr(2) = expr(2) XOR FKpi(i,2)
            CASE 2
                expr(2) = expr(2) XOR FKpi(17-i,2)
            END SELECT
            nbit = 0
            bitgrab = 0
            shiftbits = 0
            xbit = 0
            ybit = 0
            FOR t = 1 TO 8
                shiftbits = (8-t)
                bitgrab = (expr(2) SHR (shiftbits*6)) AND &H3F
                ybit = ((bitgrab SHR 5) SHL 1) OR (bitgrab AND &H1)
                xbit = (bitgrab SHR 1) AND &HF
                nbit = nbit OR (SBox(t,xbit,ybit) SHL (shiftbits*4))
            NEXT t

            STATIC AS UINTEGER P(1 TO 32) = { 16,  7, 20, 21, _
                                              29, 12, 28, 17, _
                                               1, 15, 23, 26, _
                                               5, 18, 31, 10, _
                                               2,  8, 24, 14, _
                                              32, 27,  3,  9, _
                                              19, 13, 30,  6, _
                                              22, 11,  4, 25 }
            fbit = 0
            FOR t = 1 TO 32
                fbit = DES_BitCopy(t,fbit,32,P(t),nbit,32)
            NEXT t
            RMpi(2) = LMpi(1) XOR fbit
       
            LMpi(1) = LMpi(2)
            RMpi(1) = RMpi(2)
        NEXT i
        fbit = (RMpi(1) SHL 32) OR LMpi(1)
        nbit = 0

        STATIC AS UINTEGER FinalP(1 TO 64) = { 40,  8, 48, 16, 56, 24, 64, 32, _
                                               39,  7, 47, 15, 55, 23, 63, 31, _
                                               38,  6, 46, 14, 54, 22, 62, 30, _
                                               37,  5, 45, 13, 53, 21, 61, 29, _
                                               36,  4, 44, 12, 52, 20, 60, 28, _
                                               35,  3, 43, 11, 51, 19, 59, 27, _
                                               34,  2, 42, 10, 50, 18, 58, 26, _
                                               33,  1, 41,  9, 49, 17, 57, 25 } 
        FOR i = 1 TO 64
            nbit = DES_BitCopy(i,nbit,64,FinalP(i),fbit,64)
        NEXT i

        ' CBC (Cypher Block Chaining) mode: when ENcrypting, take the value
        ' of the last encyphered block as the next Initialization Vector
        IF ed = 1 THEN IV = nbit

        ' Or when DEcrypting, XOR decyphered block with Initialization Vector
        IF ed = 2 THEN nbit = nbit XOR IV

        CypherText += HEX(nbit,16)
    NEXT z   
    CypherText = Des_HexToStr(CypherText)

    ' In DECRYPT mode, remove the padding after the decyphered text
    IF ed = 2 THEN
        PS_val = ASC(CypherText, LEN(CypherText)) ' 1 to 8
        PS = ""
        FOR i = 1 TO PS_val
            PS += CHR(PS_val)
        NEXT
        IF RIGHT(CypherText, PS_val) <> PS THEN ' Error decrypting
            CypherText = ""
        ELSE
            CypherText = LEFT(CypherText, LEN(CypherText) - PS_val)
        END IF
    END IF
    
    RETURN CypherText
   
END FUNCTION

FUNCTION DES_StrToHex (BYREF convstr AS STRING) AS STRING
    DIM AS STRING c
    c = convstr
    DIM AS UINTEGER i
    DIM AS STRING ftext
    FOR i = 1 TO LEN(c)
        ftext += HEX(ASC(MID$(c,i,1)),2)
    NEXT i
    RETURN ftext
END FUNCTION

FUNCTION DES_HexToStr (BYREF convstr AS STRING) AS STRING
    DIM AS STRING c
    c = convstr
    IF LEN(c) MOD 2 = 1 THEN c += "0"
    DIM AS UINTEGER i
    DIM AS STRING   f
    FOR i = 1 TO LEN(c) STEP 2
        f += CHR$(VAL("&H"+MID$(c,i,2)))
    NEXT i
    RETURN f
END FUNCTION

FUNCTION DES_BitCopy(BYVAL tonum AS UINTEGER, BYVAL toval AS ULONGINT, BYVAL tobits AS UINTEGER, BYVAL fromnum AS UINTEGER, BYVAL fromval AS ULONGINT, BYVAL frombits AS UINTEGER) AS ULONGINT   
    tonum = tobits-tonum
    fromnum = frombits-fromnum     
    DIM AS ULONGINT aval = (toval SHR tonum) AND &H1
    fromval = (fromval SHR fromnum) AND &H1
    RETURN ((aval XOR fromval) SHL tonum) XOR toval
END FUNCTION

SUB DES_Split56 (BYVAL mk AS ULONGINT, BYREF lh AS ULONGINT, BYREF rh AS ULONGINT)
    lh = mk SHR 28
    rh = mk SHL 28 SHR 28
END SUB

SUB DES_Split64(BYVAL mk AS ULONGINT, BYREF lh AS ULONGINT, BYREF rh AS ULONGINT)
    lh = mk SHR 32
    rh = mk SHL 32 SHR 32
END SUB
FUNCTION DES_RotateL56(BYVAL wkey AS ULONGINT, BYVAL plcs AS UINTEGER) AS ULONGINT
    SELECT CASE plcs
    CASE 1
        wkey = wkey SHL 1
        wkey = DES_BitCopy(64,wkey,64,36,wkey,64) AND &HFFFFFFF
    CASE 2
        wkey = wkey SHL 2
        wkey = DES_BitCopy(64,wkey,64,36,wkey,64)
        wkey = DES_BitCopy(63,wkey,64,35,wkey,64) AND &HFFFFFFF
    END SELECT
    RETURN wkey
END FUNCTION
