'
' This is a FreeBasic implementation of the PbeWithMD5AndDES encryption algorithm, as
' defined in RFC 2898 Password-Based Cryptography, PBSE1 encryption scheme "PKCS-5.3"
' PbeWithMD5AndDES is a very commonly used encryption/decryption algorithm in Java
'
' Usage: pbe_md5_des <action(encrypt/decrypt)> <password> <input file> [output file]
'
' User must provide a password/passphrase and a file (the content of which is the Message)
' The password is MD5-hashed N times into a Key K and an Initialization Vector IV, that
' are then used to encrypt the Message with the DES-CBC algorithm.
'
' Written by Nicolas Mougin in August 2016, open source as in the GNU GPL V3 license
'

#INCLUDE "file_utils.inc"
#INCLUDE "DES-CBC.bi"
#INCLUDE "MD5.bi"

SUB Usage()
    PRINT "Usage: pbe_md5_des <action(encrypt/decrypt)> <password> <input file> [output file]"
    END(1)
END SUB

' Parse argument #1: must be "encrypt" or "decrypt"
DIM AS LONG ed
DIM AS STRING action = LCASE(COMMAND(1))
IF action = "encrypt" THEN
    ed = 1
ELSEIF action = "decrypt" THEN
    ed = 2
ELSE
    Usage()
END IF

' Parse argument #2: password
DIM AS STRING password = COMMAND(2)

' Parse argument #3: input file
DIM AS STRING infile = COMMAND(3)
IF infile = "" THEN Usage()
IF INSTR(infile, "\") = 0 THEN infile = EXEPATH + "\" + infile
IF NOT FILEEXISTS(infile) THEN
    PRINT "Incorrect input file: " + CHR(34) + infile + CHR(34)
    END (0)
END IF

' Parse argument #4: output file
DIM AS STRING outfile = COMMAND(4)
IF outfile = "" THEN ' if not specified, set a default filename
    DIM AS LONG i = INSTR(RIGHT(infile, 5), ".")
    IF i THEN i = LEN(infile) - 6 + i ELSE i = LEN(infile)
    outfile = LEFT(infile, i) + "_" + action + MID(infile, i+1)
END IF
IF FILEEXISTS(outfile) THEN KILL outfile

' Load content of input file as the Message to be encrypted
DIM AS STRING Message = LoadFile(infile)

' 1. Select an eight-octet salt and iteration count:
DIM AS STRING salt = CHR( _
    &HA9, &H9B, &HC8, &H32, _
    &H56, &H35, &HE3, &H03 )
DIM AS LONG iterationCount = 19

' 2. Apply the PBKDF1 key derivation function:
' Calculate the derived key DK of length 16 octets by doing a
' MD5-hash of password + salt repeated 'iterationCount' times
DIM AS LONG counter
DIM AS STRING DK
DK = password + salt ' init
FOR counter = 1 TO iterationCount
    DK = MD5_raw_FromString(DK)
NEXT

' 3. Separate the derived key into an encryption
' key K, and an initialization vector IV
DIM AS STRING K_str = LEFT(DK, 8)
DIM AS STRING IV_str = RIGHT(DK, 8)
DIM AS ULONGINT K = VALLNG("&H" + DES_StrToHex(K_str))
DIM AS ULONGINT IV = VALLNG("&H" + DES_StrToHex(IV_str))

 ' 4. Cypher the Message with DES in Cyper Block Chaining mode
' (CBC) with encryption key K and initialization vector IV
DIM AS STRING cypherText = DES_Encrypt(Message, K, IV, ed)

IF LEN(cypherText) THEN
    SaveFile(outfile, cypherText)
ELSE
    PRINT "Error " + action + "inc " + CHR(34) + infile + CHR(34)
    END(1)
END IF

END(0)

